/*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 *    File: $Id: AFitBCPGenDecay.cc,v 1.5 2005/07/03 23:04:22 bevan Exp $
 * Authors:
 *    A Bevan (a.j.bevan@qmul.ac.uk)
 *
 * Copyright (C) 2009 QMUL
 *****************************************************************************/

// -- CLASS DESCRIPTION [AFit] --
// This class builds a time-dependent CP asymmetry PDF for B decays 
// produced at the Y(4S) resonance decaying into a CP eigenstate
// with oscillation amplitude coefficients S and C.

#include "AFit/AFitBCPGenDecay.hh"
#ifdef __BABARROOFIT__
#include "RooFitModels/RooBCPGenDecay.hh"
#else
#include "RooBCPGenDecay.h"
#endif

#include "TIterator.h"

ClassImp(AFitBCPGenDecay)
  ;

AFitBCPGenDecay::AFitBCPGenDecay(RooAbsReal& x, TString name)
  : 
  _avgC(name+"_C", "C", 0.0, -3.0, 3.0) ,
  _avgS(name+"_S", "S", 0.0, -3.0, 3.0) ,
  _avgMistag(name+"_avgMistag", "mistag probability", 0.0, -1.0, 1.0) ,
  _delMistag(name+"_delMistag", "delta mistag probability", 0.0, -1.0, 1.0) ,
  _mu(name+"_mu", "", 0.0, -1.0, 1.0) ,
  _tau(name+"_tau", "lifetime", 1.54, 1.0, 2.0) ,
  _dm(name+"_dm", "delta m", 0.54, 0.3, 0.7) ,
  _tag (name + "_tag", "Tag Category" ),
  _type(name + "_type", "basis type", "DoubleSided"),
  _blindStringS("GavinHenson"),
  _blindStringC("JonnyWilkinson"),
  _blindingState(name + "blindingState", "blind or unblind"),
  _resolutionModelName(name + "_resolution_model", "Resolution Model Name", "Resolution")
 {
  // Constructor
  _name     = name;

  _t = (RooRealVar*)&x;

  _blindingState.defineType("blind",1);
  _blindingState.defineType("unblind",0);
  _blindingState.setLabel("unblind");

  _tag.defineType("TagB0", 1);
  _tag.defineType("TagB0bar", -1);
  _tag.setLabel("TagB0");

  varSet.addOwned(_avgC);
  varSet.addOwned(_avgS);
  varSet.addOwned(_avgMistag);
  varSet.addOwned(_delMistag);
  varSet.addOwned(_mu);
  varSet.addOwned(_tau);
  varSet.addOwned(_dm);
  varSet.addOwned(_tag);
  varSet.addOwned(_blindingState);
  varSet.addOwned(_type);
  varSet.addOwned(_resolutionModelName);
  readDataCard();
}

AFitBCPGenDecay::~AFitBCPGenDecay() {
  delete _UnBlindC;
  delete _UnBlindS;
}

void AFitBCPGenDecay::buildPdf() {
  // Build the Pdf

  _UnBlindC = new RooUnblindOffset(_name+"UnBlindC", " unblind C", _blindStringC, 1.0, _avgC, _blindingState); 
  _UnBlindS = new RooUnblindOffset(_name+"UnBlindS", " unblind S", _blindStringS, 1.0, _avgS, _blindingState);

  _theResolutionModel = dynamic_cast<RooResolutionModel*>(globalConfig.find( _resolutionModelName.getVal() ));
  if(!_theResolutionModel) {
    cout << "  Resolution Model " << _resolutionModelName.getVal() << " does not exist" << endl;
    return;
  } else {
    cout << "  Resolution Model " << _resolutionModelName.getVal() << " has been found and will be used (address = " << _theResolutionModel << ")" <<endl;
  }

  TString thisType = _type.getVal();

  if(!thisType.CompareTo("DoubleSided", TString::kExact) ) {
    cout << "making pdf with double sided basis" << endl;
    _thePdf = new RooBCPGenDecay(_name,"time-dependent CP asymmetry PDF: double sided",*_t, _tag, _tau, _dm,
				 _avgMistag, *_UnBlindC, *_UnBlindS, _delMistag, _mu,
				 *_theResolutionModel, RooBCPGenDecay::DoubleSided);
  }
  else if(!thisType.CompareTo("SingleSided", TString::kExact) ) {
    cout << "making pdf with single sided basis" << endl;
    _thePdf = new RooBCPGenDecay(_name,"time-dependent CP asymmetry PDF: single sided",*_t, _tag, _tau, _dm,
				 _avgMistag, *_UnBlindC, *_UnBlindS, _delMistag, _mu,
				 *_theResolutionModel, RooBCPGenDecay::SingleSided);
  }
  else if(!thisType.CompareTo("Flipped", TString::kExact) ) {
    cout << "making pdf with flipped basis" << endl;
    _thePdf = new RooBCPGenDecay(_name,"time-dependent CP asymmetry PDF: flipped",*_t, _tag, _tau, _dm,
                                 _avgMistag, *_UnBlindC, *_UnBlindS, _delMistag, _mu,
                                 *_theResolutionModel, RooBCPGenDecay::Flipped);
  } else {
    cout << "unknown basis for " << ClassName() << " _type = " << _type.getVal() << " " << _thePdf << endl;
    _thePdf = 0;
  }

}

