/*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 * Author: A Bevan (a.j.bevan@qmul.ac.uk)
 *
 * Copyright (C) 2007 QMUL
 *****************************************************************************/

// -- CLASS DESCRIPTION [PDF] --
//
// Implementation of a Factory class to access multi-dimensional component 
// PDF models wrapped by an AFitComponentFactory.  The following types are 
// available:
//
//  default
//  vvpolarisation
//  composite      - a sum of component PDFs : need to debug
//
#include <iostream>

#include "TObjString.h"

#ifdef __BABARROOFIT__
#include "RooFitCore/RooAddPdf.hh"
#include "RooFitCore/RooFormulaVar.hh"
#else
#include "RooAddPdf.h"
#include "RooFormulaVar.h"
#endif

/*
 * The following includes are AFit PDFs 
 */
#include "AFit/AFitComponentFactory.hh"
#include "AFit/AFitPdfFactory.hh"
#include "AFit/AFitProdPdf.hh"

using namespace std;

AFitComponentFactory::AFitComponentFactory() :
  vvblock("")
  , _configFile("")
{
  condVarSet = 0;
  setVerbose();
  if(_debug) cout << ClassName() << _ctorstr << endl;
}

AFitComponentFactory::~AFitComponentFactory() {
}

RooAbsPdf * AFitComponentFactory::makeComp(TString name, TString type, RooArgList &discVarList) {
  RooAbsPdf * thePdf(0);
  if(condVarSet)  pdfFactory.setConditionalVariables(*condVarSet);
  pdfFactory.setConfigFile(_configFile);

  if(_debug) cout << ClassName() << " AFitProdPdf : name = " << name << endl;
  if(discVarList.getSize() == 0) {
    cout << ClassName() << " unable to build a pdf with no discriminating variables" << endl;
    return 0;
  }
  type.ToLower();

  if(type == "default") {
    /*
     * This type of PDF is a product of PDFs for the discriminating variables
     * in discVarList. 
     */
    AFitAbsPdfBuilder * theBuilder(0);
    theBuilder = pdfFactory.makePdf(name, discVarList);
    builderList.Add(theBuilder);

    thePdf = theBuilder->getPdf();

    if(_debug) cout << ClassName() << " Have built a pdf object of type " << type << " with PDF name " << name << " address = " << thePdf << endl;
  }
  else if(type == "vvpolarisation") {
    /*
     * This type of PDF has two components, a longitudinal polarisation,
     * and a transverse polarisation.  The fraction of longitudinally
     * polarised events relates the two PDFs.  This is affected by 
     * different reconstruction efficiencies by the individual polarisations
     * (NAME_effLong and NAME_effTran),
     * and so the RooAddPdf for the vvpolarisation type use an effective
     * polarisation to account for the correct weighting of longitudinal 
     * and transverse events in the fit.  The NAME_polarisationfrac variable
     * is the physical fraction of longitudinally polarised events.
     */
    TString namelong = name ; namelong += "_long";
    TString nametran = name ; nametran += "_tran";
    TString namefL = name ; namefL += "_polarisationfrac";
    TString nameEfffL = name ; nameEfffL += "_EffectivePolarisationfrac";
    TString nameEffLong = name ; nameEffLong += "_effLong";
    TString nameEffTran = name ; nameEffTran += "_effTran";

    AFitAbsPdfBuilder * theBuilderLong(0);
    AFitAbsPdfBuilder * theBuilderTran(0);
    theBuilderLong = pdfFactory.makePdf(namelong, discVarList);
    theBuilderTran = pdfFactory.makePdf(nametran, discVarList);
    builderList.Add(theBuilderLong);
    builderList.Add(theBuilderTran);

    RooAbsPdf * pdfLong = theBuilderLong->getPdf();
    RooAbsPdf * pdfTran = theBuilderTran->getPdf();
    RooRealVar * polfrac = new RooRealVar(namefL, "fL for component " + name, 0.5, -0.5, 1.5); 
    RooRealVar * efflong = new RooRealVar(nameEffLong, "efficiency for longitudinal component " + name, 0.5, 0.0, 1.0); 
    RooRealVar * efftran = new RooRealVar(nameEffTran, "efficiency for transverse component " + name, 0.5, 0.0, 1.0); 
    /*
     * Note that the epsilon = 1e-10 is added to the formula for polEff in order to make sure that the RooAbsPdf is
     * built properly when polfrac = 0.0.  If the epsilon is removed (ROOT 5.20/00) then the PDF will be zero. 
     */
    RooFormulaVar * polEff = new RooFormulaVar( nameEfffL, "effective polarisation formula", "@0/(@2*(1.0-@0)/@1+@0)+1e-10",
			      RooArgList(*polfrac, *efflong, *efftran) ); 
    RooArgSet coefSet;
    coefSet.add(*polfrac);
    coefSet.add(*efflong);
    coefSet.add(*efftran);
    coefSet.readFromFile(AFitAbsPdfBuilder::getDataCard(), 0, name);
    coefList.add(*polfrac);
    coefList.add(*efflong);
    coefList.add(*efftran);

    pdfList.add(*pdfLong);
    pdfList.add(*pdfTran);
    thePdf  = new RooAddPdf(name, "VV polarisation model " + name, *pdfLong, *pdfTran, *polEff);

    vvblock += name;
    vvblock += ",";
    cout << " ___________________________________________________________________ " << endl;
    cout << " PRINTING OUT DEBUGGING INFORMATION FOR THE VVPOLARISATION COMPONENT " << endl;
    cout << " longitudinal polarisation name = " << namelong << endl;
    cout << " RooAbsPdf / AFitAbsPdfBuilder ptr addresses = " << pdfLong << " / " << theBuilderLong << endl;
    cout << " transverse polarisation name = " << nametran << endl;
    cout << " RooAbsPdf / AFitAbsPdfBuilder ptr addresses = " << pdfTran << " / " << theBuilderTran << endl;
    coefList.Print("v");
    if(_debug) cout << ClassName() << " Have built a pdf of type " << type << " with PDF name " << name << " address = " << thePdf << endl;
    cout << " ___________________________________________________________________ " << endl;
  }
  else if(type == "composite") {
    /*
     * This type of PDF is a the sum of component PDFs, weighted by fractions.
     * for each composite - read in the name, and weight fraction, then build the PDF
     * and coefficient, adding each to a list.  Finally construct the RooAbsPdf.
     */
    RooStringVar compositePdf("compositePdf", "", "null");
    RooArgSet configSet;
    configSet.add(compositePdf);
    configSet.readFromFile(AFitAbsPdfBuilder::getDataCard(), 0, name);

    RooArgSet coefSet;
    RooArgSet compSet;
    RooArgSet compVarSet;
    TString tmpStr = compositePdf;
    TObjArray * compArr = tmpStr.Tokenize(',');
    for(Int_t i=0; i < compArr->GetEntries(); i++) {
      TString thisComp = ((TObjString*)compArr->At(i))->GetString();
      RooStringVar * thisVar = new RooStringVar(thisComp, "", "null");
      compVarSet.add(*thisVar);
    }
    compVarSet.readFromFile(AFitAbsPdfBuilder::getDataCard(), 0, name);

    TIterator * iter = compVarSet.createIterator();
    TObject * obj;
    Int_t first = 1;
    while(obj = iter->Next()) {
      RooStringVar &thisCompVar = *(RooStringVar*)obj;
      RooAbsPdf * thisComponentPdf = makeComp(thisCompVar.GetName(), thisCompVar.getVal(), discVarList );

      cout << endl << endl << " ->>>>>>>>>>>>>>>>>>> DEGUBBING COMPOSITE" << endl;
      thisComponentPdf->Print("v");
      cout << " ->>>>>>>>>>>>>>>>>>> DEGUBBING COMPOSITE" << endl;

      compSet.add(*thisComponentPdf);

      // make the coefficient to go with this component
      if(!first) {
	TString coeffName = thisCompVar.GetName();    coeffName += "Fraction";
	RooRealVar * thisCoefficient = new RooRealVar (coeffName, "coefficient", 0.0);
	coefSet.add(*thisCoefficient);
      }
      else first = 0;
    }
    coefSet.readFromFile(AFitAbsPdfBuilder::getDataCard(), 0, name);
    RooArgList coefList(coefSet);
    RooArgList compList(compSet);

    thePdf = new RooAddPdf(name, "composite model", compList, coefList);

    cout << " ->>>>>>>>>>>>>>>>>>> DEGUBBING COMPOSITE" << endl;
    thePdf->Print("V");
    cout << " ->>>>>>>>>>>>>>>>>>> DEGUBBING COMPOSITE" << endl;

    if(_debug) cout << ClassName() << " Have built an RooAddPdf object of type " << type << " with PDF name " << name << " address = " << thePdf << endl;
  } else {
    cout << ClassName() << " Unable to build a component PDF of of unknown type : type = " << type << " name = " << name << endl;
    return 0;
  }

  pdfList.add(*thePdf);
  /*
   * return the encapsulated PDF object
   */
  return thePdf;
}

void AFitComponentFactory::fixParameters(Bool_t kFix){
  /*
   * fix/free parameters for all PDFs created.
   */
  TIterator * iter = pdfList.createIterator();
  TIter next(iter);
  RooAbsPdf * obj=0;
  while (obj = (RooAbsPdf *)next()) {
    RooArgSet * compos = obj->getComponents();
    RooArgSet * pdfpar = obj->getParameters(compos);

    TIterator * pdfiter = pdfpar->createIterator();
    TIter pdfnext(pdfiter);
    TObject * pdfobj=0;
    while (pdfobj = (TObject *)pdfnext()) {
      RooRealVar * var = dynamic_cast<RooRealVar*>(obj);
      if(var) var->setConstant(kFix);
    }
  }
}

void AFitComponentFactory::writeConfigFile(ostream & out){
  /*
   * Write the PDF configuration for all known PDFs
   */
  if(vvblock != "") {
    out << "// ----------------------------------------------------------" << endl;
    out << "// " << endl;
    out << "// The following section lists the 0->11 model polarisations " << endl;
    out << "// " << endl;
    out << "// ----------------------------------------------------------" << endl;
    TObjArray * vvPdfArr = vvblock.Tokenize(',');
    for(Int_t i=0; i < vvPdfArr->GetEntries(); i++) {
      TString thisVar = ((TObjString*)vvPdfArr->At(i))->GetString();
      out << "[" << thisVar << "]" << endl;
      RooRealVar * polfrac  = (RooRealVar*)coefList.find( thisVar + "_polarisationfrac");
      RooRealVar * effL  = (RooRealVar*)coefList.find( thisVar + "_effLong");
      RooRealVar * effT  = (RooRealVar*)coefList.find( thisVar + "_effTran");

      if(polfrac) {
	out << polfrac->GetName() << " =";
	polfrac->writeToStream(out, kFALSE);
	out << endl;
      }
      if(effL) {
	out << effL->GetName() << " =";
	effL->writeToStream(out, kFALSE);
	out << endl;
      }
      if(effT) {
	out << effT->GetName() << " =";
	effT->writeToStream(out, kFALSE);
	out << endl;
      }
      out << endl << endl;
    }
  }

  out << "// __________________________________________________________" << endl;
  out << "// " << endl;
  out << "// The following section lists the configuration of the PDFs " << endl;
  out << "// (written out by the AFitPdfFactory)"<<endl; 
  out << "// " << endl;
  out << "// __________________________________________________________" << endl;
  pdfFactory.writeConfigFile(out);

  out << "// __________________________________________________________" << endl;
  out << "// " << endl;
  out << "// The following section lists pdf parameters for all shapes " << endl;
  out << "// (written out by the AFitComponentFactory)"<<endl; 
  out << "// " << endl;
  out << "// __________________________________________________________" << endl;
  TIterator * iter = builderList.MakeIterator();
  TIter next(iter);
  AFitProdPdf * obj=0;
  while (obj = (AFitProdPdf *)next()) {
    out << "// - this is an AFitProdPdf " << obj->_name << " from " << ClassName() << endl; 
    obj->writeConfigFile(out);
    out << endl<< endl;
  }
}
