/*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 * Author: A Bevan (a.j.bevan@qmul.ac.uk)
 *
 * Copyright (C) 2007 QMUL
 *****************************************************************************/

// -- CLASS DESCRIPTION [PDF] --
//
// Implementation of a Factory class to access all PDF models wrapped
// in an AFitPdfFactory.  The following PDF types are available using 
// the makePdf member function:
//
//   add:x,y,...  - Composite PDF which is the sum of x, y, ... 
//   multiply:x,y,...  - Composite PDF which is the product of x, y, ... 
//   argus          - phase space motivated distribution first used by ARGUS function.
//   breitwigner    - non-relativistic breit wigner shape
//   relbreitwigner - relativistic breit wigner shape
//   bukin          - bukin PDF
//   cbshape        - a crystal ball function
//   chebyN         - A Chebychev Polynomial of order N
//   cpdecay        - Time dependent CP model (RooBCPGenDecay) [Requires resolution function to be constructed before use!]
//   decay          - Decay lifetime model (RooDecay) [Requires resolution function to be constructed before use!]
//   bdecay         - BDecay lifetime model (RooBDecay) [Requires resolution function to be constructed before use!]
//   exp[oential]   - an exponential
//   flatte         - Flatte function
//   gaussian       - single gaussian (or asymmetric gaussian).
//   generic        - a 1D RooGenericPdf [defined by formula f(x)]
//   gounarissakurai- a Gounaris Sakurai Pdf
//   helicity       - a cos^2(x) or sin^2(x) distribution
//   1dhist         - a 1D RooHistPdf
//   1dkeys         - a 1D KEYS pdf
//   landau         - a landau function
//   novosibirsk    - a Novosibirsk function
//   psf            - a parametric step function
//   polyN          - a polynomial of order N
//   resolution     - a triple Gaussian resolution function with or without a conditional variable [specified in data card]
//   sigmoid        - a sigmoid function
//   step           - either a step function or a veto function (the step function's compliment)
//   voigtian       - a Voigtian function
//
// The following PDF types (with conditional variables) are accessible using 
// the makeConditionalPdf member function.
//
//   resolution   - a triple Gaussian resolution function with a conditional variable scaling the means and widths
//
// All PDF types are accessible by name, using a lowercase string.
//
// In addition, it is possible to make a multi-dimensional fit component 
// using the makePdf(TString name, RooArgList &discVarList) function.  
//
#include <iostream>

/*
 * The following includes are AFit PDFs 
 */
#include "AFit/AFitPdfFactory.hh"
#include "AFit/AFitAddPdf.hh"
#include "AFit/AFitMultiplyPdf.hh"
#include "AFit/AFitProdPdf.hh"
#include "AFit/AFitArgus.hh"
#include "AFit/AFitBreitWigner.hh"
#include "AFit/AFitRelBreitWigner.hh"
#include "AFit/AFitBukin.hh"
#include "AFit/AFitBCPGenDecay.hh"
#include "AFit/AFitCBShape.hh"
#include "AFit/AFitCheby.hh"
#include "AFit/AFitDecay.hh"
#include "AFit/AFitBDecay.hh"
#include "AFit/AFitFlatte.hh"
#include "AFit/AFitGaussian.hh"
#include "AFit/AFitGounarisSakurai.hh"
#include "AFit/AFitHist.hh"
#include "AFit/AFitKeys.hh"
#include "AFit/AFitNovosibirskShape.hh"
#include "AFit/AFitPoly.hh"
#include "AFit/AFitParametricStepFunction.hh"
#include "AFit/AFitExponential.hh"
#include "AFit/AFitLandau.hh"
#include "AFit/AFitVoigtianShape.hh"
#include "AFit/AFitResolution.hh"
#include "AFit/AFitSigmoid.hh"
#include "AFit/AFitStepFunction.hh"
#include "AFit/AFitHelicity.hh"
#include "AFit/AFitGeneric.hh"

using namespace std;

AFitPdfFactory::AFitPdfFactory() {
  condVarSet = 0;
  setVerbose();
  if(_debug) cout << ClassName() << _ctorstr << endl;
}

AFitPdfFactory::~AFitPdfFactory() {
}

AFitAbsPdfBuilder * AFitPdfFactory::makePdf(TString name, RooArgList &discVarList) {
  /*
   * Factory interface to the AFitProdPdf class.  The return type is an AFitAbsPdfBuilder
   * pointer.  The arguments passed to this function are:
   *
   *   name        = the PDF name
   *   discVarList = RooArgList of discriminating variables
   *
   ***************************************************************************************
   * Example: make a pdf component called 'component' which is three dimensional:
   *
   *  RooRealVar x("x", "", -1, 3);
   *  RooRealVar y("y", "", 0, 3);
   *  RooRealVar q("q", "", -2, 3);
   *  AFitAbsPdfBuilder::setDataCard("datacardProdPdf.txt");
   *  RooArgList ll;
   *  ll.add(x);
   *  ll.add(y);
   *  ll.add(q);
   *
   *  AFitPdfFactory fact;
   *  AFitAbsPdfBuilder * componentPDF = fact.makePdf("component", ll);
   *
   * where the datacard snippet required to build the PDF is of the form:
   *
   * [component]
   * component_x_type = gaussian
   * component_y_type = poly2
   * component_q_type = landau
   *
   * The [component] block defines the shapes for each discriminating variable, in this case
   * the x, y and q varibles.  The shapes defined by this block have names 'component_x,y,q'
   * and are themselves defined in the following lines 
   *
   * [component_x]
   * component_xmean = 1.5 +/- 0.1
   * component_xwidth = 0.5 +/- 0.1
   *
   * [component_y]
   * component_y_p0 = 1.0 +/- 0.1
   * component_y_p1 = 1.0 +/- 0.1
   * component_y_p2 = 0.5 +/- 0.1
   *
   * [component_q]
   * component_qmean = 1.5 +/- 0.1
   * component_qwidth = 0.5 +/- 0.1
   *
   * Using this method of AFitPfdFactory, it is possible to build a multidimensional PDF 
   * component from a list of discriminating variables, and a datacard configuration.
   */
  AFitAbsPdfBuilder * isBuilt = (AFitAbsPdfBuilder *)builderList.FindObject(name);
  if(isBuilt) {
    cout << "A pdf with name " << name << " already exists - returning this" << endl;
    return isBuilt;
  }

  AFitAbsPdfBuilder * theBuilder(0);
  if(_debug) cout << ClassName() << " AFitProdPdf : name = " << name << endl;
  if(discVarList.getSize() == 0) {
    cout << ClassName() << " unable to build an AFitProdPdf with no discriminating variables" << endl;
    return 0;
  }

  theBuilder = new AFitProdPdf(discVarList, name);
  if(condVarSet) ((AFitProdPdf*)theBuilder)->setConditionalVariables(*condVarSet);
  ((AFitProdPdf*)theBuilder)->setConfigFile(_configFile);

  if(_debug) cout << ClassName() << " Have built an AFitProdPdf object with PDF name " << name << " address = " << theBuilder << endl;

  builderList.Add(theBuilder);
  /*
   * return the encapsulated PDF object
   */
  return theBuilder;

}

AFitAbsPdfBuilder * AFitPdfFactory::makeConditionalPdf(TString name, TString type, RooAbsReal & var, RooAbsReal & conditionalvar) {
  /*
   * Factory interface to the AFit pdf builders.  The return type is an AFitAbsPdfBuilder
   * pointer.  The arguments passed to this function are:
   *
   *   name            = the PDF name
   *   type            = lowercase string corresponding to the AFitAbsPdfBuilder type to create.  In the case
   *                     of and AFitAddPdf, the name is 'add:pdf1,pdf2,...,pdfN', where pdfi are the PDF
   *                     types to be made 
   *   var             = the discriminating variable of the PDF being created.
   *   conditionalvar  = the conditional variable of the PDF being created (e.g. sigma(deltat) for a time resolution function).
   */
  AFitAbsPdfBuilder * isBuilt = (AFitAbsPdfBuilder *)builderList.FindObject(name);
  if(isBuilt) {
    cout << "A pdf with name " << name << " already exists - returning this" << endl;
    return isBuilt;
  }

  AFitAbsPdfBuilder * theBuilder(0);
  type.ToLower();
  type.ReplaceAll(" ", "");
  if(_debug) cout << ClassName() << " PDF type = " << type << " name = " << name << endl;
  /*
   * Make the AFitAbsPdfBuilder based on the requested type
   */
  if(!type.CompareTo("resolution")) {
    if(_debug) cout << "Resolution Function called " << name << endl;

    theBuilder = new AFitResolution(&var, &conditionalvar, name);
  }

  if(_debug) cout << ClassName() << " Have built encapsulated PDF object of type " << type << " with PDF name " << name << " address = " << theBuilder << endl;

  builderList.Add(theBuilder);

  /*
   * return the encapsulated PDF object
   */
  return theBuilder;
}

AFitAbsPdfBuilder * AFitPdfFactory::makePdf(TString name, TString type, RooAbsReal & var) {
  /*
   * Factory interface to the AFit pdf builders.  The return type is an AFitAbsPdfBuilder
   * pointer.  The arguments passed to this function are:
   *
   *   name = the PDF name
   *   type = lowercase string corresponding to the AFitAbsPdfBuilder type to create.  In the case
   *          of and AFitAddPdf, the name is 'add:pdf1,pdf2,...,pdfN', where pdfi are the PDF
   *          types to be made 
   *   var  = the discriminating variable of the PDF being created.
   */

  AFitAbsPdfBuilder * isBuilt = (AFitAbsPdfBuilder *)builderList.FindObject(name);
  if(isBuilt) {
    cout << "A pdf with name " << name << " already exists - returning this" << endl;
    return isBuilt;
  }

  AFitAbsPdfBuilder * theBuilder(0);
  type.ToLower();
  type.ReplaceAll(" ", "");
  if(_debug) cout << ClassName() << " PDF type = " << type << " name = " << name << endl;
  /*
   * Make the AFitAbsPdfBuilder based on the requested type
   */

  if(type.Contains("multiply:")) {
    TString types = type;
    types.ReplaceAll("multiply:", "");
    if(_debug) cout << "Composite PDF called " << name << " of type " << types << endl;
    theBuilder = new AFitMultiplyPdf(var, name, types);
    if(condVarSet) ((AFitMultiplyPdf*)theBuilder)->setConditionalVariables(*condVarSet);
    ((AFitMultiplyPdf*)theBuilder)->setConfigFile(_configFile);
  }
  else if(type.Contains("add:")) {
    TString types = type;
    types.ReplaceAll("add:", "");
    if(_debug) cout << "Composite PDF called " << name << " of type " << types << endl;
    theBuilder = new AFitAddPdf(var, name, types);
    if(condVarSet) ((AFitAddPdf*)theBuilder)->setConditionalVariables(*condVarSet);
    ((AFitAddPdf*)theBuilder)->setConfigFile(_configFile);
  }
  else if(!type.CompareTo("argus")) {
    if(_debug) cout << "Argus called " << name << endl;
    theBuilder = new AFitArgus(var, name);
  }
  else if(!type.CompareTo("breitwigner")) {
    if(_debug) cout << "Non-relativistic Breit Wigner called " << name << endl;
    theBuilder = new AFitBreitWigner(var, name);
  }
  else if(!type.CompareTo("relbreitwigner")) {
    if(_debug) cout << "relativistic Breit Wigner called " << name << endl;
    theBuilder = new AFitRelBreitWigner(var, name);
  }
  else if(!type.CompareTo("bukin")) {
    if(_debug) cout << "Bukin function called " << name << endl;
    theBuilder = new AFitBukin(var, name);
  }
  else if(!type.CompareTo("cbshape")) {
    if(_debug) cout << "CB shape called " << name << endl;
    theBuilder = new AFitCBShape(var, name);
  }  else if(type.Contains("cheby")) {
    cout << "here" << endl;
    TString tmp = type;
    tmp.ReplaceAll("cheby", "");
    Int_t order = tmp.Atoi();
    if(_debug) cout << "Chebychev Polynomial called " << name << " of order " << order << endl;
    theBuilder = new AFitCheby(var, name, order);
    cout << theBuilder << endl;
  }  else if(!type.CompareTo("cpdecay")) {
    RooStringVar tmpVar(name + "_resolution_model", "Resolution model name", "");
    RooArgSet tmpSet(tmpVar);
    tmpSet.readFromFile(_configFile, 0, name);

    // build the resolution model PDF before building the decay PDF
    AFitAbsPdfBuilder * bld = makePdf(tmpVar.getVal(), "resolution", var);
    RooAbsPdf * pdf = bld->getPdf();

    if(_debug) cout << "AFitBCPGenDecay CP decay model shape called " << name << endl;
    theBuilder = new AFitBCPGenDecay(var, name);
  } else if(!type.CompareTo("decay")) {
    RooStringVar tmpVar(name + "_resolution_model", "Resolution model name", "");
    RooArgSet tmpSet(tmpVar);
    tmpSet.readFromFile(_configFile, 0, name);

    // build the resolution model PDF before building the decay PDF
    AFitAbsPdfBuilder * bld = makePdf(tmpVar.getVal(), "resolution", var);
    RooAbsPdf * pdf = bld->getPdf();

    if(_debug) cout << "AFitDecay model shape called " << name << endl;
    theBuilder = new AFitDecay(var, name);
  }
  else if(!type.CompareTo("bdecay")) {
    RooStringVar tmpVar(name + "_resolution_model", "Resolution model name", "");
    RooArgSet tmpSet(tmpVar);
    tmpSet.readFromFile(_configFile, 0, name);

    // build the resolution model PDF before building the decay PDF
    AFitAbsPdfBuilder * bld = makePdf(tmpVar.getVal(), "resolution", var);
    RooAbsPdf * pdf = bld->getPdf();

    if(_debug) cout << "AFitBDecay model shape called " << name << endl;
    theBuilder = new AFitBDecay(var, name);
  }
  else if(!type.CompareTo("exp") || !type.CompareTo("exponential")) {
    if(_debug) cout << "Exponential shape called " << name << endl;
    theBuilder = new AFitExponential(var, name);
  }
  else if(!type.CompareTo("flatte")) {
    if(_debug) cout << "flatte function called " << name << endl;
    theBuilder = new AFitFlatte(var, name);
  }
  else if(!type.CompareTo("gaussian")) {
    if(_debug) cout << "gaussian called " << name << endl;
    theBuilder = new AFitGaussian(var, name);
  }
  else if(!type.CompareTo("generic")) {
    if(_debug) cout << "1D Generic PDF shape called " << name << endl;
    theBuilder = new AFitGeneric(var, name);
  }
  else if(!type.CompareTo("gounarissakurai")) {
    if(_debug) cout << "Gounaris Sakurai PDF shape called " << name << endl;
    theBuilder = new AFitGounarisSakurai(var, name);
  }
  else if(!type.CompareTo("helicity")) {
    if(_debug) cout << "Helicity shape called " << name << endl;
    theBuilder = new AFitHelicity(var, name);
  }
  else if(!type.CompareTo("1dhist")) {
    if(_debug) cout << "1D Hist PDF shape called " << name << endl;
    theBuilder = new AFitHist(var, name);
  }
  else if(!type.CompareTo("1dkeys")) {
    if(_debug) cout << "1D KEYS PDF shape called " << name << endl;
    theBuilder = new AFitKeys(var, name);
  }
  else if(!type.CompareTo("landau")) {
    if(_debug) cout << "Landau shape called " << name << endl;
    theBuilder = new AFitLandau(var, name);
  }
  else if(!type.CompareTo("novosibirsk")) {
    if(_debug) cout << "Novosibirsk shape called " << name << endl;
    theBuilder = new AFitNovosibirskShape(var, name);
  }
  else if(type.Contains("poly")) {
    TString tmp = type;
    tmp.ReplaceAll("poly", "");
    Int_t order = tmp.Atoi();
    if(_debug) cout << "Polynomial called " << name << " of order " << order << " (type = " << type << ")" << endl;
    theBuilder = new AFitPoly(var, name, order);
  }
  else if(!type.CompareTo("psf")) {
    if(_debug) cout << "Parametric Step Function shape called " << name << endl;
    theBuilder = new AFitParametricStepFunction(var, name);
  }
  else if(!type.CompareTo("resolution")) {
    if((condVarSet) && condVarSet->getSize()) {
      RooStringVar tmpVar(name + "_conditional_variable", "conditional variable name", "");
      RooArgSet tmpSet(tmpVar);
      tmpSet.readFromFile(_configFile, 0, name);
      RooRealVar * condvar = (RooRealVar*)condVarSet->find(tmpVar.getVal());
      if(_debug) cout << "Resolution Function called " << name << " [will make with the conditional variable " << condvar->GetName()  << "] "<< endl;
      theBuilder = new AFitResolution(&var, condvar, name);
    } else {
      cout << "cond var set = " << condVarSet << " " << _configFile << endl;
      if(_debug) cout << "Resolution Function called " << name << " [without a conditional variable] "<< endl;
      theBuilder = new AFitResolution(&var, 0, name);
    }
  }
  else if(!type.CompareTo("sigmoid")) {
    if(_debug) cout << "Sigmoid shape called " << name << endl;
    theBuilder = new AFitSigmoid(var, name);
  }
  else if(!type.CompareTo("step")) {
    if(_debug) cout << "Step function shape called " << name << endl;
    theBuilder = new AFitStepFunction(var, name);
  }
  else if(!type.CompareTo("voigtian")) {
    if(_debug) cout << "Voigtian shape called " << name << endl;
    theBuilder = new AFitVoigtianShape(var, name);
  } else {
    cout << ClassName() << " ERROR: Unable to build encapsulated PDF object of type " << type << " with PDF name " << name << endl;
    assert(0);
  }

  if(_debug) cout << ClassName() << " Have built encapsulated PDF object of type " << type << " with PDF name " << name << " address = " << theBuilder << endl;

  builderList.Add(theBuilder);
  /*
   * return the encapsulated PDF object
   */
  return theBuilder;
}

void AFitPdfFactory::writeConfigFile(ostream & out) {
  /*
   * Write the PDF configuration for all known PDFs
   */
  cout << ClassName() << " -- starting to loop over pdfs made by AFitPdfFactory" << endl;
  TIterator * iter = builderList.MakeIterator();
  TIter next(iter);
  AFitAbsPdfBuilder * obj=0;
  while (obj = (AFitAbsPdfBuilder *)next()) {
    cout << endl << endl << "******************************************" 
	 << "Writing composites: AFitAbsPdfBuilder " << obj->_name << endl; 
    cout << "Will print out the PDFs for an AFitAbsPdfBuilder object" << endl;

    obj->writeDataCard(out);
    out << endl<< endl;
  }
}

