/*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 *    File: $Id: AFitProdPdf.cc,v 1.5 2005/07/03 23:04:22 bevan Exp $
 * Authors:
 *    A Bevan (a.j.bevan@qmul.ac.uk)
 *
 * Copyright (C) 2007 QMUL
 *****************************************************************************/

// -- CLASS DESCRIPTION [AFit] --
//
//  AFitProdPdf(RooArgList& depList, TString name) :
//
// This class builds a composite PDF (RooProdPdf) based on a given list of 
// list of dependant variables, and types.
//
// The RooProdPdf built by this class is of the form
//
//        PDF_1(x1) * PDF_2(x2) * ... * PDF_N(xN)
// 
// Example:
//
// Computer resources dictate the limit of the number of composite PDFs that
// can be constructed.  
//

#include "AFit/AFitProdPdf.hh"
#include "AFit/AFitAddPdf.hh"
#ifdef __BABARROOFIT__
#include "RooFitCore/RooProdPdf.hh"
#else
#include "RooProdPdf.h"
#endif

#include "TString.h"

ClassImp(AFitProdPdf)
  ;

AFitProdPdf::AFitProdPdf(RooArgList& depList, TString name) :
  _depList(depList)
  , _typeList("typeList")
  , _configFile("AFitProdPdf")
{
  // Constructor
  _name  = name;
  nTypes       = depList.getSize();

  TString dependents = "";
  for(int i = 0; i < nTypes; i++) {
    RooAbsReal & thisDepVar = (RooAbsReal&)_depList[i];
    TString thisName = _name; thisName += "_"; thisName += thisDepVar.GetName();
    thisName += "_";  thisName += "type";

    cout << thisName << " " << i << endl;

    _typeList.add( *(new RooStringVar(thisName, "pdf type string", "uninitilised")) );

    dependents+=thisDepVar.GetName();
    if(i != (nTypes-1)) dependents+=",";
  }
  cout << " Number of dimensions in PDF " << name << " = " << nTypes << " (" << dependents << ")" << endl;

  varSet.add(_typeList);

  readDataCard();
}

AFitProdPdf::~AFitProdPdf() {
  cout << ClassName() << ":destructor to implement DELETE THE INSTANTIATED PDFS!" << endl;
}

void AFitProdPdf::buildPdf() {
  // Build the Pdf
  if(condVarSet) fact.setConditionalVariables(*condVarSet);
  fact.setConfigFile(_configFile);

  cout << " TO FINISH IMPLEMENTEING  " << ClassName() << endl;
  _depList.Print("v");
  _typeList.Print("v");
  for(int i = 0; i < nTypes; i++) {
    RooAbsReal & thisDepVar = (RooAbsReal&)_depList[i];
    RooStringVar & thisTypeVar = (RooStringVar&)_typeList[i];
    cout << "This dependant variable = " << thisDepVar.GetName() << " type = " <<thisTypeVar.getVal() << endl;
    TString thisType = thisTypeVar.getVal();
    TString thisName = _name; thisName += "_"; thisName += thisDepVar.GetName();
    AFitAbsPdfBuilder * bld = (AFitAbsPdfBuilder*)fact.makePdf(thisName, thisType, thisDepVar);
    bldList.Add(bld);
    pdfList.add(*bld->getPdf());
  }
  pdfList.Print("v");

  _thePdf = new RooProdPdf(_name, "Composite PDF from AFitProdPdf", pdfList);
}

void AFitProdPdf::writeConfigFile(ostream & out) {
  /*
   * Write the PDF configuration for all known PDFs
   */
  TIterator * iter = bldList.MakeIterator();
  TIter next(iter);
  AFitAbsPdfBuilder * obj=0;
  while (obj = (AFitAbsPdfBuilder *)next()) {
    out << "// " << ClassName() << " PDF name, title = " << obj->GetName()<< ", " << obj->_name << ", "<< obj->GetTitle() << endl;
    obj->writeDataCard(out);
    out << endl<< endl;
  }
}
