 /*
 * class AFitStatTools
 *
 *   A. Bevan, a.j.bevan@qmul.ac.uk
 *   Copyright (C) Queen Mary, University of London
 *   Sun Oct  5 15:08:42 BST 2008
 */


#include "AFit/AFitStatTools.hh"
#include <iostream>

#include "TString.h"
#include "TMath.h"
#include "TObjString.h"
#include "TObjArray.h"

using namespace RooFit;

ClassImp(AFitStatTools)
  ;

AFitStatTools::AFitStatTools()
{
}

Double_t AFitStatTools::binomialError(Int_t p, Int_t N){
  /*
   * Compute the binomial error from a sample of N events where
   * p events have been selected.
   */
  if((N==0) || (p == N)) return 0.0;
  Double_t eff = (Double_t)p/(Double_t)N;
  Double_t err = sqrt((1.0 - eff)*eff/(Double_t)N);
  return err;
}

Double_t AFitStatTools::chi2(const Int_t N, const Double_t * x, const Double_t * xhat, const Double_t * sigma){
  /*
   * Compute the chi2 sum for input data.
   */
  Double_t chi2(0.0);
  for(int i=0; i < N; i++) {
    chi2 += (x[i] - xhat[i])*(x[i] - xhat[i])/sigma[i]/sigma[i];
  }

  return chi2;
}

Double_t AFitStatTools::chi2prob(const Int_t N, const Double_t * x, const Double_t * xhat, const Double_t * sigma, const Int_t nu){
  /*
   * Compute the chi2 probability, given the input data and number of degrees of freedom.
   */
  Double_t thechi2 = chi2(N, x, xhat, sigma);
  Double_t chi2pr = TMath::Prob(thechi2, nu);

  return chi2pr;
}

void AFitStatTools::correlation(TTree * tree, TString variables, Bool_t kType) {
  if(kType) spearmans_rank_correlation(tree, variables);
  else pearsons_correlation(tree, variables);
}

void AFitStatTools::pearsons_correlation(TTree * tree, TString variables) {
  /*
   * Given a comma separated list of branch names for the data contained in the tree,
   * compute and print out the correlation matrix.
   */
  TObjArray * varArr = variables.Tokenize(',');
  const Int_t dimN = varArr->GetEntries();
  Double_t rho[dimN][dimN];

  for(int i=0; i < dimN; i++)  rho[i][i] = 1.0;


  for(int i=0; i < dimN; i++) 
    for(int j=0; j < dimN; j++) {
      if(i>j) {
	// compute the correlation
	TString svar1 = ((TObjString*)varArr->At(i))->GetString();
	TString svar2 = ((TObjString*)varArr->At(j))->GetString();
	tree->ResetBranchAddresses();
	Double_t var1;
	Double_t var2;
	tree->SetBranchAddress(svar1, &var1);
	tree->SetBranchAddress(svar2, &var2);

	Double_t mean1(0.0);
	Double_t mean2(0.0);
	Double_t mean1sqr(0.0);
	Double_t mean2sqr(0.0);
	Double_t mean12(0.0);
	Double_t s1(0.0);
	Double_t s2(0.0);
	Double_t s12(0.0);

	for(int k=0; k < tree->GetEntries(); k++) {
	  tree->GetEntry(k);
	  mean1 += var1;
	  mean2 += var2;
	  mean1sqr += var1*var1;
	  mean2sqr += var2*var2;
	  mean12 += var1*var2;
	} /* loop over entries in the tree */

	mean1 /= tree->GetEntries();
	mean2 /= tree->GetEntries();
	mean1sqr /= tree->GetEntries();
	mean2sqr /= tree->GetEntries();
	mean12 /= tree->GetEntries();

	s1 = sqrt( mean1sqr - mean1*mean1 );
	s2 = sqrt( mean2sqr - mean2*mean2 );
	s12 = mean12 -mean1*mean2;

	if((s1 != 0) && (s2 != 0) ) {
	  rho[i][j] = s12/s1/s2;
	  rho[j][i] = s12/s1/s2;
	} else {
	  rho[i][j] = errval;
	  rho[j][i] = errval;
	}
      } /* select one half of the matrix to calculate */
    } /* looping over variable combinations in the correlation matrix */

  tree->ResetBranchAddresses();

  cout << "___________________________________________________________" << endl;
  cout << " Results from AFitStatTools::pearsons_correlation for the variables " << endl;
  cout << " \t" << variables << endl;
  cout << " Pearsons Correlation matrix follows: " << endl;

  for(int i=0; i < dimN; i++) {
    for(int j=0; j < dimN; j++) {
      cout.width(7);
      cout << setprecision(6) << rho[i][j] << " \t";
    }
    cout << endl;
  }
  cout << "___________________________________________________________" << endl;
}

void AFitStatTools::spearmans_rank_correlation(TTree * tree, TString variables) {
  /*
   * Given a comma separated list of branch names for the data contained in the tree,
   * compute and print out the pearsons rank correlation matrix.
   *
   * Works with continuous variable data, and does not deal with equally 
   * ranked data points.  This limitation should not be an issue for 
   * correlations calculated for real data sets.
   *
   */
  TObjArray * varArr = variables.Tokenize(',');
  const Int_t dimN = varArr->GetEntries();
  Double_t rho[dimN][dimN];

  for(int i=0; i < dimN; i++) rho[i][i] = 1.0;

  for(int i=0; i < dimN; i++) 
    for(int j=0; j < dimN; j++) {
      if(i<j) {
	TString svar1 = ((TObjString*)varArr->At(i))->GetString();
	TString svar2 = ((TObjString*)varArr->At(j))->GetString();
	tree->ResetBranchAddresses();
	Double_t var1;
	Double_t var2;
	tree->SetBranchAddress(svar1, &var1);
	tree->SetBranchAddress(svar2, &var2);

	// compute the correlation
	const Int_t n = tree->GetEntries();
	Double_t x[n];
	Double_t y[n];
	Int_t rx[n];
	Int_t ry[n];

	// get the data in local arrays in order to rank x against y 
	for(Int_t k=0; k < n; k++) {
	  tree->GetEntry(k);
	  x[k] = var1;
	  y[k] = var2;
	}

	// rank the data in order to calculate the correlation
	bubbleSort(n, x, y, rx, ry);
	
	// calculate spearman's rank correlation for these data.
	Double_t rho_s = 0.0;
	for(Int_t k=0; k < n; k++) {
	  //cout << " (x, y) = (" << x[k] <<", " << y[k] << ")" << endl;
	  rho_s += (rx[k] - ry[k])*(rx[k] - ry[k]);
	}
	rho[i][j] = 1.0 - 6*rho_s/( n * (n*n - 1.0));
	rho[j][i] = rho[i][j];

	/*
	  cout << " Ranked data:" << endl;
	  cout << "IQ        TV         r(IQ)        r(TV)        d            d^2" << endl; 
	  for(int k=0; k<n; k++) {
	  cout << x[rx[k]-1] << "\t" << y[rx[k]-1] << "\t" << rx[k] << "\t" << ry[k] << "\t" << (rx[k] - ry[k]) << "\t" << (rx[k] - ry[k])*(rx[k] - ry[k])  << endl;
	  }
	*/

      } /* select one half of the matrix to calculate */
    } /* looping over variable combinations in the correlation matrix */
  
  cout << "___________________________________________________________" << endl;
  cout << " Results from AFitStatTools::spearmans_rank_correlation for the variables " << endl;
  cout << " \t" << variables << endl;
  cout << " Spearmans Rank Correlation matrix follows: " << endl;
  
  for(int i=0; i < dimN; i++) {
    for(int j=0; j < dimN; j++) {
      cout.width(7);
      cout << setprecision(6) << rho[i][j] << " \t";
    }
    cout << endl;
  }
  cout << "___________________________________________________________" << endl;
}


void AFitStatTools::bubbleSort(const Int_t Narr, Double_t * arr1, Double_t * arr2, Int_t * rank1, Int_t * rank2)
{
  /*
   * Sort the arrays in terms of ascending values [order is preserved in arr1 and arr2
   * but not in rank1 and rank2 - which is the ranked indices of the lowest to higest
   * values.  These indicies need to be re-ordered into numerical rankings
   */
  Int_t index1[Narr];
  Int_t index2[Narr];
  bubbleSort(Narr, arr1, rank1, index1);
  bubbleSort(Narr, arr2, rank2, index2);
  Int_t newRank1[Narr];
  Int_t newRank2[Narr];
  Double_t newArr1[Narr];
  Double_t newArr2[Narr];

  /*
  cout << "____________________ bubble done for individual arrays: ___________ " << endl;
  cout << "ARRAY 1" << endl;
  for(Int_t i=0; i < Narr; i++) {
    cout << i << " \t " << arr1[index1[i]] << " " << rank1[i] << " " << index1[i] << endl;
  }
  cout << "ARRAY 2" << endl;
  for(Int_t i=0; i < Narr; i++) {
    cout << i << " \t " << arr2[index2[i]] << " " << rank2[i] << " " << index2[i] << endl;
  }
  */

  // now sort the arrays of data according to the rank of the first array
  for(Int_t i=0; i < Narr; i++) {
    newRank1[i] = rank1[i];
    newArr1[i] = arr1[index1[i]];
    newArr2[i] = arr2[index1[i]];
  }

  // now have to associate the sorted arr2 values with their rank.
  for(Int_t i=0; i < Narr; i++) 
    for(Int_t j=0; j < Narr; j++) 
      if(newArr2[i] == arr2[index2[j]]) newRank2[i] = j + 1;

  for(Int_t i=0; i < Narr; i++) {
    arr1[i] = newArr1[i];
    arr2[i] = newArr2[i];
    rank1[i] = newRank1[i];
    rank2[i] = newRank2[i];
  }
  //cout << "____________________ arrays have been sorted now for correlation calculation: ___________ " << endl;
  //for(Int_t i=0; i < Narr; i++) {
  //  cout << i << " \t " << arr1[i] << " " << rank1[i] <<" " << " " << arr2[i] << " " << rank2[i] <<endl;
  //}
}

void AFitStatTools::bubbleSort(const Int_t Narr, const Double_t * arr, Int_t * rank, Int_t * index)
{
  /*
   * the array arr is sorted from lowest to highest values, with 
   * the indicies being given by the rank array.  rank[0] is the index of the
   * lowest value of arr, etc.
   */
  Double_t localArr[Narr];
  Int_t localRank[Narr];

  for(Int_t iEl = 0; iEl < Narr; iEl++) {
    localArr[iEl] = arr[iEl];
    localRank[iEl] = iEl;
  }

  // sort the data, and keep track of the original index
  for(Int_t iEl = 0; iEl < Narr; iEl++)    {
    for(Int_t iEl2 = Narr-1; iEl2 > iEl; --iEl2)	{
      if(localArr[iEl2-1] > localArr[iEl2])	{
	double tmp        = localArr[iEl2-1];
	localArr[iEl2-1]  = localArr[iEl2];
	localArr[iEl2]    = tmp;
	
	Int_t tmp2        = localRank[iEl2-1];
	localRank[iEl2-1] = localRank[iEl2];
	localRank[iEl2]   = tmp2;
      }
    }
  }

  //  cout << "__________________________"<<endl;
  //cout << "Sorted data:" << endl;
  for(Int_t iEl = 0; iEl < Narr; iEl++)    {
    rank[iEl] = iEl+1;
    index[iEl] = localRank[iEl];
    //cout << " value, rank, idx " << localArr[iEl] << " " << rank[iEl] << " "<< index[iEl] << endl;
  }
  //  cout << "__________________________"<<endl;
}
