using namespace RooFit;
 
void cpfit() {
  cout << "AFit Example: cpfit.cc" << endl;
  cout << "__________________________________________________________________"<<endl;
  // the discriminating variable is deltat
  RooRealVar deltat("deltat", "dt", -10.0, 10.0);

  // the error on delta t is a conditional variable (to be used later)
  RooRealVar deltatErr("deltatErr", "sdt", 0.0, 2.50);

  /*
   * The resolution function, called SigReso depends on both
   * the discriminating variable, and the conditional variable.
   *
   * The parmeter values are set by hand, and fixed to be
   * constant for this example.
   */
  AFitResolution resoBld(&deltat, &deltatErr, "SigReso");
  resoBld.scaleCoreMean.setLabel("yes");
  resoBld.scaleCoreWidth.setLabel("yes");
  resoBld.scaleTailMean.setLabel("yes");
  resoBld.scaleTailWidth.setLabel("yes");
  resoBld.sigMean1.setVal(0.0);
  resoBld.sigSig1.setVal(1.1);
  resoBld.sigCoreFrac.setVal(1.0);
  resoBld.sigTailFrac.setVal(0.0);
  resoBld.sigMean1.setConstant();
  resoBld.sigSig1.setConstant();
  resoBld.sigMean2.setConstant();
  resoBld.sigSig2.setConstant();
  resoBld.sigMean3.setConstant();
  resoBld.sigSig3.setConstant();
  resoBld.sigCoreFrac.setConstant();
  resoBld.sigTailFrac.setConstant();

  RooAbsPdf * resopdf = resoBld.getPdf();

  cout << "resolution model address = " << resopdf << endl;
  cout << "_________________________________________________" << endl;
  cout << "Will now try and use resolution in AFitBCPGenDecay model" << endl;

  /*
   * The parameters for the AFitBCPGenDecay model include the physical
   * observables S and C, as well as the mistag parameters, mutag and 
   * resolution model name (must match an existing resolution model).
   *
   * The parameters tau and dm are the B0 lifetime and mixing frequencies,
   * respectively.  It is not possible to simultaneously reliably extract dm 
   * at the same time as S and C as the S and C amplitudes depend 
   * on deltam oscillations in the time-evolution.
   *
   * The normal state for S and C is 'unblind'.  If the blindingState
   * is set to be blind, then the true values of S and C are hidden.
   * by the RooUnblindOffset class.
   */
  AFitBCPGenDecay mymodel(deltat, "SignalModel");
  mymodel._avgC.setVal(0.0);
  mymodel._avgS.setVal(0.7);
  mymodel._avgMistag.setVal(0.1);
  mymodel._delMistag.setVal(0.0);
  mymodel._type = "DoubleSided";
  mymodel._blindingState.setLabel("unblind");
  mymodel._resolutionModelName = "SigReso";
  mymodel._avgMistag.setConstant();
  mymodel._delMistag.setConstant();
  mymodel._mu.setConstant();
  mymodel._tau.setConstant();
  mymodel._dm.setConstant();

  RooAbsPdf * pdf = mymodel.getPdf();

  RooCategory * tagCategory = &mymodel._tag;
  tagCategory->Print("v");

  /*
   * Now that the PDF model has been made, construct a prototype
   * data set that contains the B and Bbar tagged event information.
   */
  const Int_t Ngen = 10000;
  // make prototype
  RooDataSet protoB("protoB", "", RooArgSet(*tagCategory));
  for(int i=0; i < Ngen; i++) protoB.add(RooArgSet(*tagCategory));
  tagCategory->setIndex(-1);
  RooDataSet protoBbar("protoBbar", "", RooArgSet(*tagCategory));
  for(int i=0; i < Ngen; i++) protoBbar.add(RooArgSet(*tagCategory));

  /*
   * We also need to make a prototype for the deltat error quantity.
   * This is done using a Landau distribution, and then merged with
   * the flavour tag category prototype that we made above.
   */
  cout << "Will now add sigma(deltat) to proto type - use a landau"<<endl;
  AFitLandau land(deltatErr, "land");
  land.mean.setVal(1.1);
  land.sigma.setVal(0.2);

  RooAbsPdf * landPDF = land.getPdf();
  RooDataSet * protoDTE = landPDF->generate(RooArgSet(deltatErr), Ngen);

  protoB.merge(protoDTE);
  protoBbar.merge(protoDTE);


  /*
   * Now the merged prototype data can be used to generate data samples
   * for fitting and plotting.
   */
  cout << "-----------------------------------" << endl;
  RooDataSet * dataB = pdf->generate(RooArgSet(deltat), protoB, Ngen);
  RooDataSet * dataBbar = pdf->generate(RooArgSet(deltat), protoBbar, Ngen);

  // append the two data sets for fitting, and for making the asymmetry
  // plot of all of the data.
  RooArgSet * aset = (RooArgSet*)dataB.get(0);
  RooDataSet data("data", "", *aset);
  data.append(*dataB); 
  data.append(*dataBbar); 

  RooFitResult * r = pdf->fitTo(data, "trm");

  /*
   * Now use the generated data to make plots.
   */
  RooPlot * frameB0 = deltat.frame();
  RooPlot * frameB0bar = deltat.frame();
  RooPlot * frameAsym = deltat.frame();

  dataB->plotOn(frameB0);
  dataBbar->plotOn(frameB0bar);

  pdf->plotOn(frameB0,  ProjWData(*dataB), LineColor(kBlue));
  pdf->plotOn(frameB0bar, ProjWData(*dataBbar), LineColor(kBlue) );

  data->plotOn(frameAsym, Asymmetry(*tagCategory) );
  pdf->plotOn(frameAsym, Asymmetry(*tagCategory), LineColor(kRed), ProjWData(*protoDTE) );

  TCanvas can("can", "");
  can.Divide(1,3);
  can.cd(1);
  frameB0->Draw();
  can.cd(2);
  frameB0bar->Draw();
  can.cd(3);
  frameAsym->Draw();

  can.Print("fig/cpfit.eps");
  cout << " ---------------- done -----------------------" << endl;
}

